@@ -29,7 +29,7 @@ export interface Diagnostic {
   message: string
   /// An optional custom rendering function that displays the message
   /// as a DOM node.
-  renderMessage?: () => Node
+  renderMessage?: (view: EditorView) => Node
   /// An optional array of actions that can be taken on this
   /// diagnostic.
   actions?: readonly Action[]
@@ -45,7 +45,7 @@ export interface Action {
   apply: (view: EditorView, from: number, to: number) => void
 }
 
-type DiagnosticFilter = (diagnostics: readonly Diagnostic[]) => Diagnostic[]
+type DiagnosticFilter = (diagnostics: readonly Diagnostic[], state: EditorState) => Diagnostic[]
 
 interface LintConfig {
   /// Time to wait (in milliseconds) after a change before running
@@ -61,6 +61,12 @@ interface LintConfig {
   /// Filter applied to a set of diagnostics shown in a tooltip. No
   /// tooltip will appear if the empty set is returned.
   tooltipFilter?: null | DiagnosticFilter
+  /// Can be used to control what kind of transactions cause lint
+  /// hover tooltips associated with the given document range to be
+  /// hidden. By default any transactions that changes the line
+  /// around the range will hide it. Returning null falls back to this
+  /// behavior.
+  hideOn?: (tr: Transaction, from: number, to: number) => boolean | null
 }
 
 interface LintGutterConfig {
@@ -88,7 +94,7 @@ class LintState {
     let markedDiagnostics = diagnostics
     let diagnosticFilter = state.facet(lintConfig).markerFilter
     if (diagnosticFilter)
-      markedDiagnostics = diagnosticFilter(markedDiagnostics)
+      markedDiagnostics = diagnosticFilter(markedDiagnostics, state)
 
     let ranges = Decoration.set(markedDiagnostics.map((d: Diagnostic) => {
       // For zero-length ranges or ranges covering only a line break, create a widget
@@ -99,7 +105,8 @@ class LintState {
         }).range(d.from)
         : Decoration.mark({
           attributes: {class: "cm-lintRange cm-lintRange-" + d.severity + (d.markClass ? " " + d.markClass : "")},
-          diagnostic: d
+          diagnostic: d,
+          inclusive: true
         }).range(d.from, d.to)
     }), true)
     return new LintState(ranges, panel, findDiagnostic(ranges))
@@ -117,8 +124,11 @@ function findDiagnostic(diagnostics: DecorationSet, diagnostic: Diagnostic | nul
 }
 
 function hideTooltip(tr: Transaction, tooltip: Tooltip) {
+  let from = tooltip.pos, to = tooltip.end || from
+  let result = tr.state.facet(lintConfig).hideOn(tr, from, to)
+  if (result != null) return result
   let line = tr.startState.doc.lineAt(tooltip.pos)
-  return !!(tr.effects.some(e => e.is(setDiagnosticsEffect)) || tr.changes.touchesRange(line.from, line.to))
+  return !!(tr.effects.some(e => e.is(setDiagnosticsEffect)) || tr.changes.touchesRange(line.from, Math.max(line.to, to)))
 }
 
 function maybeEnableLint(state: EditorState, effects: readonly StateEffect<unknown>[]) {
@@ -178,7 +188,7 @@ export function diagnosticCount(state: EditorState) {
   return lint ? lint.diagnostics.size : 0
 }
 
-const activeMark = Decoration.mark({class: "cm-lintRange cm-lintRange-active"})
+const activeMark = Decoration.mark({class: "cm-lintRange cm-lintRange-active", inclusive: true})
 
 function lintTooltip(view: EditorView, pos: number, side: -1 | 1) {
   let {diagnostics} = view.state.field(lintState)
@@ -193,7 +203,7 @@ function lintTooltip(view: EditorView, pos: number, side: -1 | 1) {
   })
 
   let diagnosticFilter = view.state.facet(lintConfig).tooltipFilter
-  if (diagnosticFilter) found = diagnosticFilter(found)
+  if (diagnosticFilter) found = diagnosticFilter(found, view.state)
 
   if (!found.length) return null
 
@@ -288,7 +298,7 @@ const lintPlugin = ViewPlugin.fromClass(class {
     } else {
       this.set = false
       let {state} = this.view, {sources} = state.facet(lintConfig)
-      Promise.all(sources.map(source => Promise.resolve(source(this.view)))).then(
+      if (sources.length) Promise.all(sources.map(source => Promise.resolve(source(this.view)))).then(
         annotations => {
           let all = annotations.reduce((a, b) => a.concat(b))
           if (this.view.state.doc == state.doc)
@@ -323,16 +333,17 @@ const lintPlugin = ViewPlugin.fromClass(class {
   }
 })
 
-const lintConfig = Facet.define<{source: LintSource, config: LintConfig},
+const lintConfig = Facet.define<{source: LintSource | null, config: LintConfig},
                                 Required<LintConfig> & {sources: readonly LintSource[]}>({
   combine(input) {
     return {
-      sources: input.map(i => i.source),
+      sources: input.map(i => i.source).filter(x => x != null) as readonly LintSource[],
       ...combineConfig(input.map(i => i.config), {
         delay: 750,
         markerFilter: null,
         tooltipFilter: null,
-        needsRefresh: null
+        needsRefresh: null,
+        hideOn: () => null,
       }, {
         needsRefresh: (a, b) => !a ? b : !b ? a : u => a(u) || b(u)
       })
@@ -342,9 +353,10 @@ const lintConfig = Facet.define<{source: LintSource, config: LintConfig},
 
 /// Given a diagnostic source, this function returns an extension that
 /// enables linting with that source. It will be called whenever the
-/// editor is idle (after its content changed).
+/// editor is idle (after its content changed). If `null` is given as
+/// source, this only configures the lint extension.
 export function linter(
-  source: LintSource,
+  source: LintSource | null,
   config: LintConfig = {}
 ): Extension {
   return [
@@ -380,7 +392,7 @@ function renderDiagnostic(view: EditorView, diagnostic: Diagnostic, inPanel: boo
   let keys = inPanel ? assignKeys(diagnostic.actions) : []
   return elt(
     "li", {class: "cm-diagnostic cm-diagnostic-" + diagnostic.severity},
-    elt("span", {class: "cm-diagnosticText"}, diagnostic.renderMessage ? diagnostic.renderMessage() : diagnostic.message),
+    elt("span", {class: "cm-diagnosticText"}, diagnostic.renderMessage ? diagnostic.renderMessage(view) : diagnostic.message),
     diagnostic.actions?.map((action, i) => {
       let fired = false, click = (e: Event) => {
         e.preventDefault()
@@ -702,7 +714,7 @@ class LintGutterMarker extends GutterMarker {
 
     let diagnostics = this.diagnostics
     let diagnosticsFilter = view.state.facet(lintGutterConfig).tooltipFilter
-    if (diagnosticsFilter) diagnostics = diagnosticsFilter(diagnostics)
+    if (diagnosticsFilter) diagnostics = diagnosticsFilter(diagnostics, view.state)
 
     if (diagnostics.length)
       elt.onmouseover = () => gutterMarkerMouseOver(view, elt, diagnostics)
@@ -795,7 +807,7 @@ const lintGutterMarkers = StateField.define<RangeSet<GutterMarker>>({
       if (effect.is(setDiagnosticsEffect)) {
         let diagnostics = effect.value
         if (diagnosticFilter)
-          diagnostics = diagnosticFilter(diagnostics || [])
+          diagnostics = diagnosticFilter(diagnostics || [], tr.state)
         markers = markersForDiagnostics(tr.state.doc, diagnostics.slice(0))
       }
     }
